﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

// Fanorona AI-player game implementation
// kenrick@uaa.alaska.edu
// 11/2009
// 
// Use source at your own risk
// TODO: Proper documentation

namespace Fanornator
{
    public partial class Form1 : Form
    {
        private static int CELLSIZE = 50;
        private Board b = new Board();
        private int whoseturn = Board.WHITE;
        // Start time for move
        private DateTime startTime;
        public const int MAXTIME = 20;      // max # of seconds to make a move

        public Form1()
        {
            InitializeComponent();
        }

        private void pboxBoard_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            // Draw grid
            for (int x=0; x<8; x++)
                for (int y = 0; y < 4; y++)
                {
                    int xc = (x * CELLSIZE) + (CELLSIZE / 2);
                    int yc = (y * CELLSIZE) + (CELLSIZE / 2);

                    /*
                     * Test the connections made correctly by drawing the board
                     */
                    if (Board.isConnected(x,y,x+1,y+1))
                        g.DrawLine(Pens.Black, xc, yc, xc + CELLSIZE, yc + CELLSIZE);
                    if (Board.isConnected(x,y+1, x+1,y))
                        g.DrawLine(Pens.Black, xc, yc + CELLSIZE, xc + CELLSIZE, yc);
                    if (Board.isConnected(x,y, x+1,y))
                        g.DrawLine(Pens.Black, xc, yc, xc + CELLSIZE, yc);
                    if (Board.isConnected(x, y, x, y+1))
                        g.DrawLine(Pens.Black, xc, yc, xc, yc + CELLSIZE);
                    if (Board.isConnected(x, y+1, x + 1, y+1))
                        g.DrawLine(Pens.Black, xc, yc + CELLSIZE, xc + CELLSIZE, yc + CELLSIZE);
                    if (Board.isConnected(x+1, y+1, x + 1, y))
                        g.DrawLine(Pens.Black, xc + CELLSIZE, yc + CELLSIZE, xc + CELLSIZE, yc);

                    /*
                    if ((((x % 2) == 0) && ((y % 2) == 0))
                        ||
                         (((x % 2) == 1) && ((y % 2) == 1)))
                         g.DrawLine(Pens.Black, xc, yc, xc + CELLSIZE, yc + CELLSIZE);
                    else if ((((x % 2) == 0) && ((y % 2) == 1)) 
                            ||
                             (((x % 2) == 1) && ((y % 2) == 0)))
                        g.DrawLine(Pens.Black, xc, yc + CELLSIZE, xc + CELLSIZE, yc);
                    g.DrawLine(Pens.Black, xc, yc, xc + CELLSIZE, yc);
                    g.DrawLine(Pens.Black, xc, yc, xc, yc + CELLSIZE);
                    g.DrawLine(Pens.Black, xc, yc + CELLSIZE, xc + CELLSIZE, yc + CELLSIZE);
                    g.DrawLine(Pens.Black, xc + CELLSIZE, yc + CELLSIZE, xc + CELLSIZE, yc);
                     */
                }
            // Draw pieces
            for (int x = 0; x < 9; x++)
                for (int y = 0; y < 5; y++)
                {
                    int color = b.board[x, y];
                    int xc = (x * CELLSIZE) + (CELLSIZE / 2);
                    int yc = (y * CELLSIZE) + (CELLSIZE / 2);
                    if (color == Board.BLACK)
                    {
                        g.FillEllipse(Brushes.Black, xc - 10, yc - 10, 20, 20);
                    }
                    else if (color == Board.WHITE)
                    {
                        g.FillEllipse(Brushes.White, xc - 10, yc - 10, 20, 20);
                        g.DrawEllipse(Pens.Black, xc - 10, yc - 10, 20, 20);
                    }
                }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            txtMove.Focus();
            ShowMoveList();
        }

        private void ShowMoveList()
        {
            Console.WriteLine("Turn: " + whoseturn);
            // Show list of all moves to the console
            List<Move> movelist = b.MoveList(whoseturn);
            foreach (Move m in movelist)
            {
                Console.WriteLine(m.ToString());
            }
            Console.WriteLine();
        }

        private void btnMove_Click(object sender, EventArgs e)
        {
            lblMessage.Text = "";
            string s = txtMove.Text.Trim().ToUpper();
            int x1, y1, x2, y2;
            char direction;
            direction = s[0];
            x1 = Convert.ToInt32(s[1] - '0');
            y1 = 4 - Convert.ToInt32(s[2] - '0');       // Convert to internal format
            x2 = Convert.ToInt32(s[3] - '0');
            y2 = 4 - Convert.ToInt32(s[4] - '0');       // COnvert to internal format
            if (direction == 'M')      // Single move, zugzwang
            {
                Move m = new Move(x1, y1, x2, y2);
                if (!b.ValidMove(m, whoseturn))
                {
                    lblMessage.Text = "Invalid Move.  Try again.";
                }
                else
                {
                    lstMoves.Items.Add(s);
                    lstMoves.SetSelected(lstMoves.Items.Count - 1, true);
                    b.MakeMove(m);
                    UpdateTurn();
                }
            }
            else
            {
                // List of captures separated by a space
                string[] arr = s.Split(" ".ToCharArray()); 
                Move m = new Move();
                foreach (string capture in arr)
                {
                    direction = capture[0];
                    // Translate to internal format
                    if (direction == 'C')
                        direction = 'A';                        
                    x1 = Convert.ToInt32(capture[1] - '0');
                    y1 = 4- Convert.ToInt32(capture[2] - '0');     // Convert
                    x2 = Convert.ToInt32(capture[3] - '0');
                    y2 = 4- Convert.ToInt32(capture[4] - '0');     // Convert
                    m.AddCapture(x1, y1, x2, y2, direction);
                }
                if (!b.ValidMove(m, whoseturn))
                {
                    lblMessage.Text = "Invalid Move.  Try again.";
                }
                else
                {
                    lstMoves.Items.Add(s);
                    lstMoves.SetSelected(lstMoves.Items.Count - 1, true);
                    b.MakeMove(m);
                    UpdateTurn();
                }
            }

            ShowMoveList();

            pboxBoard.Invalidate();
            txtMove.Text = "";
            txtMove.Focus();

            // Check for game over
            int result = b.gameOver();
            if (result == Board.BLACK)
            {
                lblMessage.Text = "BLACK WINS!";
                txtMove.Enabled = false;
                btnMove.Enabled = false;
            }
            else if (result == Board.WHITE)
            {
                lblMessage.Text = "WHITE WINS!";
                txtMove.Enabled = false;
                btnMove.Enabled = false;
            }
        }

        private void UpdateTurn()
        {
            whoseturn = b.OpponentColor(whoseturn);
            if (whoseturn == Board.BLACK)
                lblWhoseTurn.Text = "Black's Turn.";
            else
                lblWhoseTurn.Text = "White's Turn.";
        }

        private void txtMove_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 13)
            {
                btnMove_Click(sender, e);
                e.Handled = true;
            }
        }

        private void quitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void helpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            frmHelp hlp = new frmHelp();
            hlp.ShowDialog();
        }

        private Move ComputerMove()
        {
            lblMessage.Text = "Computer thinking...";
            Cursor.Current = Cursors.WaitCursor;
            Application.DoEvents();

            //Make random move
            //List<Move> moveList = b.getMoveList(whoseturn);
            //int i = rand.Next(moveList.Count);
            //Move m = moveList[i];
            Move m = getAIMove(whoseturn);
            Cursor.Current = Cursors.Default;
            lblMessage.Text = "";
            Application.DoEvents();
            return m;
        }

        private Move getAIMove(int whichplayer)
        {
            int stop_depth = Convert.ToInt32(this.cboLookahead.Text);
            int max_depth = 2;  // Start at depth 2 and go higher if we have time
            Move move = null;
            Move bestmove = null;
            int val;
            startTime = DateTime.Now;       // Start the timer
            bool outOfTime = false;
            // Start visible timer
            lblTime.Text = "0";
            timer1.Enabled = true;


            // First try depth of 1 to see if it can win, if so, take it
            val = MaxValue(b, whichplayer, whichplayer, -999999, 999999, 0, 1, ref move, false);
            if (val == Int32.MaxValue)
            {
                timer1.Enabled = false;
                return move;
            }
            else
            {
                // Loop to deeper depths until we run out of time or win
                bestmove = move;
                while ((!outOfTime) && (max_depth <= stop_depth))
                {
                    val = MaxValue(b, whichplayer, whichplayer, -999999, 999999, 0, max_depth, ref move, true);
                    DateTime t2 = DateTime.Now;
                    TimeSpan ts = t2 - startTime;
                    int diffSeconds = ts.Seconds;
                    if (diffSeconds < MAXTIME)
                    {
                        // Finished the search in time, remember best move and search one deeper
                        bestmove = new Move(move);
                        Console.WriteLine("Finished depth " + max_depth);
                        max_depth++;
                    }
                    else
                    {
                        // if we ran out of time throw away the results of this search as
                        // it may be incomplete
                        outOfTime = true;
                    }
                }
            }
            timer1.Enabled = false;
            return bestmove;
        }

        private int MaxValue(Board b, int originalplayer, int whoseturn,
                             int alpha_max, int beta_min,
                             int cur_depth, int max_depth, ref Move theMove, bool addmove)
        {
            List<Move> moveList;
            if (cur_depth == max_depth)
                return Heuristic(b, originalplayer);

            // See if out of time
            DateTime t2 = DateTime.Now;
            TimeSpan ts = t2 - startTime;
            int diffSeconds = ts.Seconds;
            if (diffSeconds >= MAXTIME)
            {
                return Heuristic(b, originalplayer);
            }

            moveList = b.MoveList(whoseturn);            

            // We're screwed if no moves possible? 
            if (moveList.Count == 0)
                return alpha_max;

            if (addmove && (moveList != null))  // Add best move from lower level search to front of list
            {
                moveList.Insert(0, theMove);
            }
            for (int i = 0; i < moveList.Count; i++)
            {
                Application.DoEvents();
                if ((i == 0) && (cur_depth == 0))
                    theMove = moveList[0];   // Pick first move as best move by default
                Board newBoard = new Board(b);
                newBoard.MakeMove(moveList[i]);
                int value;
                int result = newBoard.gameOver();
                if (result == originalplayer)
                {
                    // Winning move
                    value = Int32.MaxValue;
                }
                else if (result != Board.BLANK)
                {
                    // Losing move
                    value = -1 * Int32.MaxValue;
                }
                else
                {
                    int opponent  = b.OpponentColor(whoseturn);
                    value = MinValue(newBoard, originalplayer, opponent, alpha_max, beta_min,
                                     cur_depth + 1, max_depth);
                }
                if (value > alpha_max)
                {
                    alpha_max = value;
                    theMove = moveList[i];
                }
                if (alpha_max >= beta_min) return (alpha_max);
            }
            return alpha_max;
        }

        private int MinValue(Board b, int originalplayer, int whoseturn,
                             int alpha_max, int beta_min,
                             int cur_depth, int max_depth)
        {
            List<Move> moveList;
            if (cur_depth == max_depth)
                return Heuristic(b, originalplayer);

            // See if out of time
            DateTime t2 = DateTime.Now;
            TimeSpan ts = t2 - startTime;
            int diffSeconds = ts.Seconds;
            if (diffSeconds >= MAXTIME)
            {
                return Heuristic(b, originalplayer);
            }

            moveList = b.MoveList(whoseturn);

            // We're screwed if no moves possible? 
            if (moveList.Count == 0)
                return beta_min;

            for (int i = 0; i < moveList.Count; i++)
            {
                Application.DoEvents();
                Board newBoard = new Board(b);
                newBoard.MakeMove(moveList[i]);
                int value;
                int result = newBoard.gameOver();
                if (result == originalplayer)
                {
                    // Winning move
                    value = Int32.MaxValue;
                }
                else if (result != Board.BLANK)
                {
                    // Losing move
                    value = -1 * Int32.MaxValue;
                }
                else
                {
                    int opponent = b.OpponentColor(whoseturn);

                    Move dummyMove = null;
                    value = MaxValue(newBoard, originalplayer, opponent, alpha_max, beta_min,
                                     cur_depth + 1, max_depth, ref dummyMove, false);
                }
                if (value < beta_min)
                {
                    beta_min = value;
                }
                if (alpha_max >= beta_min) return (beta_min);
            }
            return beta_min;
        }

        // Very simple, just (Mypieces - Hispieces)*100 + (44 - mypieces - hispieces)
        private int Heuristic(Board b, int originalplayer)
        {
            int mypieces = 0, hispieces = 0;
            for (int x=0; x<9; x++)
                for (int y = 0; y < 5; y++)
                {
                    if (b.board[x, y] == originalplayer)
                        mypieces++;
                    else if (b.board[x, y] == Board.BLANK)
                    {
                        // nothing
                    }
                    else
                    {
                        hispieces++;
                    }
                }
            return (mypieces - hispieces)*100 + (44 - mypieces - hispieces);
        }

        private void button1_Click(object sender, EventArgs e)
        {
            ShowMoveList();
        }

        private void btnCompMove_Click(object sender, EventArgs e)
        {
            lblMessage.Text = "";
            Move m = ComputerMove();
            string s = m.ToString();

            lstMoves.Items.Add(s);
            lstMoves.SetSelected(lstMoves.Items.Count - 1, true);
            b.MakeMove(m);
            UpdateTurn();
                           
            ShowMoveList();

            pboxBoard.Invalidate();
            txtMove.Text = "";
            txtMove.Focus();

            // Check for game over
            int result = b.gameOver();
            if (result == Board.BLACK)
            {
                lblMessage.Text = "BLACK WINS!";
                txtMove.Enabled = false;
                btnMove.Enabled = false;
                btnCompMove.Enabled = false;
            }
            else if (result == Board.WHITE)
            {
                lblMessage.Text = "WHITE WINS!";
                txtMove.Enabled = false;
                btnMove.Enabled = false;
                btnCompMove.Enabled = false;
            }
            Application.DoEvents();
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            int num = Convert.ToInt32(lblTime.Text);
            num++;
            lblTime.Text = num.ToString();
            timer1.Enabled = true;
            Application.DoEvents();
        }


    }
}
